import urllib.request
import json
import subprocess
import os
import ssl
import logging
import xml.etree.ElementTree as ET

logging.basicConfig()
log = logging.getLogger()

### Globals ###
GALAXY_URL = "http://galaxy.quadram.ac.uk/galaxy/u/nabil/w/"
WORKFLOW_PATH = (
    "/usr/users/QIB_fr005/alikhan/code/irida/src/main/resources/ca/corefacility/bioinformatics/"
    "irida/model/workflow/analysis/type/workflows/"
)
WORKFLOW_PROP = (
    "/usr/users/QIB_fr005/alikhan/code/irida/src/main/resources/ca/corefacility/bioinformatics/"
    "irida/config/workflows.properties"
)
PIPELINES = [
    {
        "galaxy_url": GALAXY_URL + "assemblyannotation-2" + "/json",
        "irida_type": "ASSEMBLY_ANNOTATION",
        "irida_name": "AssemblyAnnotation",
        "irida_version": "0.7.7",
        "workflow_path": WORKFLOW_PATH,
        "uuid": "4ab671c3-ebfb-4cae-82a6-2facb3d09f23",
    },
    {
        "galaxy_url": GALAXY_URL + "phylogeneticsmaster" + "/json",
        "irida_type": "PHYLOGENOMICS",
        "irida_name": "Phylogenetics",
        "irida_version": "0.1.4",
        "workflow_path": WORKFLOW_PATH,
        "multiple": True,
        "uuid": "4b6685e3-db29-4064-9364-409bed00ea05",
    },
]

## Do things ##
# First read in config for workflows
final_config = {}
with open(WORKFLOW_PROP) as wf:
    for j in wf.readlines():
        if len(j.split("=")) == 2:
            final_config[j.split("=")[0]] = j.split("=")[1].strip()
for pipe in PIPELINES:
    # Download JSON from Galaxy
    ctx = ssl.create_default_context()
    ctx.check_hostname = False
    ctx.verify_mode = ssl.CERT_NONE
    with urllib.request.urlopen(pipe["galaxy_url"], context=ctx) as url:
        data = json.loads(url.read().decode())
        for x, y in data["steps"].items():
            if y["label"] == "reference":
                y["tool_state"] = '{"name": "reference"}'
        with open("{irida_name}.json".format(**pipe), "w") as outfile:
            json.dump(data, outfile)
        # Generate IRIDA files from galaxy workflow
        process_ga = (
            "java -jar lib/irida-wf-ga2xml-1.0.0-SNAPSHOT-standalone.jar "
            "-i {irida_name}.json "
            "-o {workflow_path} "
            "-W {irida_version} "
            "-t {irida_type} "
            "-n {irida_name}".format(**pipe)
        )
        if pipe.get("multiple"):
            process_ga += " --multi-sample "
        subprocess.call(process_ga, shell=True)
        # Check and clean up errors in IRIDA workflow output
        workflow_xml_path = os.path.join(
                WORKFLOW_PATH,
                pipe["irida_name"],
                pipe["irida_version"],
                "irida_workflow.xml",
            )
        workflow_xml = ET.parse(workflow_xml_path)
        workflow_root = workflow_xml.getroot()
        # CHECK XML: has sequence Reads Paired specified
        for sq_field in workflow_root.iterfind('inputs/sequenceReadsPaired'):
            sq_field.text = 'sequence_reads_paired'
        # CHECK XML: Get workflow ID
        for id_field in workflow_root.iterfind('id'):
            final_config["irida.workflow.default.{}".format(pipe["irida_type"])] = id_field.text
        # Write XML
        workflow_xml.write(open(workflow_xml_path, 'w'), encoding='unicode')
        # CHECK JSON: Make sure galaxy json has the name specified.
        ga_json = os.path.join(
            WORKFLOW_PATH,
            pipe["irida_name"],
            pipe["irida_version"],
            "irida_workflow_structure.ga",
        )
        x = json.load(open(ga_json))
        # CHECK JSON: Make sure input labels are specified
        input_exists = False
        for step_num, step in x['steps'].items():
            if step.get('label') == 'sequence_reads_paired':
                if len(step['inputs']) < 1 or not step['inputs'][0].get('name'):
                    log.warning('inputs field for input reads not specified')
                step['inputs'] = [dict(description="", name="sequence_reads_paired")]
                input_exists = True
        # CHECK JSON: Make sure inputs are specified
        if not input_exists:
            log.exception('Read input is not specified correctly. No sequence_read_paired component found')
        with open(ga_json, "w") as ga_file:
            ga_file.write(json.dumps(x, indent=4, sort_keys=True))


# Finally reformat workflow config, to have new hashes for the updated pipelines
with open(WORKFLOW_PROP, "w") as wf:
    for k, v in final_config.items():
        wf.write("{}={}\n".format(k, v))



import subprocess
import os
import csv
from multiprocessing import Process, Queue
import logging
from subprocess import Popen
import copy
import shutil
import time

log = logging.getLogger()


class FetchReads:
    def __init__(
        self,
        Sample_name,
        Accession,
        File_Forward,
        File_Reverse,
        aspera_path,
        aspera_key,
        output_dir,
        fastq_dump_path,
        Project_ID,
    ):
        self.acc = Accession
        self.sample_name = Sample_name
        self.aspera_path = aspera_path
        self.aspera_key = aspera_key
        self.output_dir = output_dir
        self.fastq_dump_path = fastq_dump_path
        self.r1 = File_Forward
        self.r2 = File_Reverse
        self.project_id = Project_ID

    def run(self):
        if not os.path.exists(self.output_dir):
            os.mkdir(self.output_dir)
        if self.acc:
            self.r1 = os.path.join(self.output_dir, self.acc + "_1.fastq.gz")
            self.r2 = os.path.join(self.output_dir, self.acc + "_2.fastq.gz")
            if True:
                # if os.path.exists(self.r1):
                log.debug("Exists, skipping " + self.r1)
            else:
                ascp_command = [
                    os.path.join(self.aspera_path, "ascp"),
                    "-i",
                    self.aspera_key,
                    "-k",
                    "1",
                    "-T",
                    "-l200m",
                ]
                ftp_path = "anonftp@ftp.ncbi.nlm.nih.gov:/sra/sra-instant/reads/ByRun/sra/{:.3}/{:.6}/{}/{}.sra".format(
                    self.acc, self.acc, self.acc, self.acc
                )
                ascp_command += [ftp_path, self.output_dir]
                sra_file = os.path.join(self.output_dir, self.acc + ".sra")
                if not os.path.exists(sra_file):
                    log.debug("Running " + " ".join(ascp_command))
                    subprocess.call(ascp_command)
                fastq_dump_command = [
                    os.path.join(self.fastq_dump_path, "fastq-dump"),
                    "--gzip",
                    "--split-files",
                    os.path.join(self.output_dir, self.acc + ".sra"),
                    "--outdir",
                    os.path.join(self.output_dir),
                ]
                log.debug("Running " + " ".join(fastq_dump_command))
                subprocess.call(fastq_dump_command)
                log.info("Downloaded {}".format(self.sample_name))
                os.remove(sra_file)
        else:
            # Handling local files
            ori_r1 = self.r1
            ori_r2 = self.r2
            self.r1 = os.path.join(self.output_dir, self.sample_name + "_1.fastq.gz")
            self.r2 = os.path.join(self.output_dir, self.sample_name + "_2.fastq.gz")
            log.info("Copying {}".format(self.sample_name))
            if not os.path.exists(self.r1):
                shutil.copy(ori_r1, self.r1)
            if not os.path.exists(self.r2):
                shutil.copy(ori_r2, self.r2)
        return dict(
            sample_name=self.sample_name,
            project_id=self.project_id,
            r1=self.r1,
            r2=self.r2,
        )


def worker(q, output_queue):
    while True:
        obj = q.get()
        output_queue.put(obj.run())
        if q.empty():
            log.info("spanner {}".format(q.qsize()))
            break


def prepare_sample(
    sample_file, aspera_path, aspera_key, fastq_dump_path, output_dir, num_workers=2
):
    upload_queue = Queue()
    output_queue = Queue()
    workers = []
    for x in range(num_workers):
        upload_proc = Process(target=worker, args=(upload_queue, output_queue))
        upload_proc.start()
        workers.append(upload_proc)
    config = dict(
        aspera_path=aspera_path,
        aspera_key=aspera_key,
        fastq_dump_path=fastq_dump_path,
        output_dir=output_dir,
    )
    # Try to finish off read files already saved locally.
    for x in sorted(
        [x for x in csv.DictReader(open(sample_file))],
        key=lambda i: i["File_Forward"],
        reverse=True,
    ):
        if x.get("Sample_name"):
            if x.get("Accession") or (x.get("File_Forward") and x.get("File_Reverse")):
                y = copy.deepcopy(config)
                y.update(x)
                upload_queue.put(FetchReads(**y))
                log.debug("Queued: {}".format(x.get("Sample_name")))
            else:
                log.warning(
                    "Sample {} does not have enough input, skipping".format(
                        x.get("Sample_name")
                    )
                )
        else:
            log.warning("Sample has no name, skipping")
    log.info("{} Samples Queued.".format(upload_queue.qsize()))
    time.sleep(1)
    for x in workers:
        x.join()
    log.info("Read files ready!")
    all_results = [output_queue.get() for x in workers]
    upload_irida(all_results, output_dir)
    log.info("Samples uploaded.")


def upload_irida(all_results, output_dir):
    irida_sample_file = os.path.join(output_dir, "SampleList.csv")
    with open(irida_sample_file, "w") as temp_sheet:
        temp_sheet.write("[Data]\n")
        temp_sheet.write("Sample_Name,Project_ID,File_Forward,File_Reverse\n")
        for res in all_results:
            if os.path.exists(res["r1"]) and os.path.exists(res["r2"]):
                temp_sheet.write(
                    "{},{},{},{}\n".format(
                        res["sample_name"],
                        res["project_id"],
                        os.path.basename(res["r1"]),
                        os.path.basename(res["r2"]),
                    )
                )
            else:
                log.error(" Path {} or {} do not exist ".format(res["r1"], res["r2"]))
        temp_sheet.close()
        log.info("Uploading to IRIDA")
        stderr = b"ERROR"
        count = 1
        while "ERROR" in stderr.decode("utf-8"):
            if count > 1:
                log.warning("{}".format(stderr))
                time.sleep(1)
            log.info("Uploading  to IRIDA attempt {}".format(count))
            p = Popen(
                [
                    "python",
                    "/usr/users/QIB_fr005/alikhan/code/irida-uploader/upload_run.py",
                    "-c",
                    "/usr/users/QIB_fr005/alikhan/code/qi_irida_utils/config.conf",
                    output_dir,
                ],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            stdout, stderr = p.communicate()
            count += 1



# -*- coding: utf-8 -*-

"""Console script for qi_irida_utils."""
import sys
import click
import os
from irida_workflow import prepare_sample
import logging

logging.basicConfig()
log = logging.getLogger()


@click.group()
def cli1(args=None):
    pass


@cli1.command()
@click.option(
    "--aspera_path",
    default=os.path.join(os.path.expanduser("~"), ".aspera/connect/bin/"),
    help="Path of ascp binary",
)
@click.option(
    "--fastqdump_path",
    default=os.path.join(os.path.expanduser("~"), "app/sratoolkit/bin/"),
    help="Path of fastq-dump binary",
)
@click.option(
    "--aspera_key",
    default=os.path.join(
        os.path.expanduser("~"), ".aspera/connect/etc/asperaweb_id_dsa.openssh"
    ),
    help="Path of ascp key",
)
@click.option(
    "--sample_file",
    default=os.path.join(os.path.expanduser("~"), "test_file.csv"),
    help="Path of sample file",
)
@click.option("--output_dir", default="/tmp/irwork", help="Output directory")
@click.option("--verbose", default=True, help="Verbose output")
def upload(aspera_path, fastqdump_path, aspera_key, sample_file, output_dir, verbose):
    """Uploads sequencing data (SRA or local) to IRIDA"""
    if verbose:
        log.setLevel(logging.DEBUG)
    if not os.path.exists(output_dir):
        os.mkdir(output_dir)
    prepare_sample(sample_file, aspera_path, aspera_key, fastqdump_path, output_dir)
    return 0


@cli1.command()
def download(args=None):
    """Console script for qi_irida_utils."""
    click.echo("Not Implemented!")
    return 0


@cli1.command()
def sync_workflow(args=None):
    """Syncronise workflow between galaxy and local IRIDA environment"""
    click.echo("Not Implemented!")
    return 0


cli = click.CommandCollection(sources=[cli1])

if __name__ == "__main__":
    sys.exit(cli())  # pragma: no cover



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `qi_irida_utils` package."""

import pytest
from click.testing import CliRunner
import sys
import os
my_path = os.path.abspath(os.path.dirname(__file__))
sys.path.insert(0, os.path.join(my_path, '../qi_irida_utils'))
print(sys.path)
from cli import cli

@pytest.fixture
def response():
    """Sample pytest fixture.

    See more at: http://doc.pytest.org/en/latest/fixture.html
    """
    # import requests
    # return requests.get('https://github.com/audreyr/cookiecutter-pypackage')


def test_content(response):
    """Sample pytest test function with the pytest fixture as an argument."""
    # from bs4 import BeautifulSoup
    # assert 'GitHub' in BeautifulSoup(response.content).title.string


def test_command_line_interface():
    """Test the CLI."""
    runner = CliRunner()
    help_result = runner.invoke(cli, ['--help'])
    assert help_result.exit_code == 0
    assert 'Show this message and exit.' in help_result.output

def test_connection():
    """Assumes localhost, tests Oauth"""
